package trees.bintree;
import java.io.*; 

class BTBasicNode implements Comparable<BTBasicNode> {

    BTBasicNode left;
    BTBasicNode right;

    public BTBasicNode() {
	left = null;
	right = null;
    }

    public BTBasicNode getLeft() {
	return left;
    }

    public void setLeft(BTBasicNode theNode) {
	left = theNode;
    }

    public BTBasicNode getRight() {
	return right;
    }

    public void setRight(BTBasicNode theNode) {
	right = theNode;
    }

    public int compareTo(BTBasicNode theNode) {
	return 1;
    }

    public String toString() {
	return " (null) ";
    }

    public BTBasicNode find(BTBasicNode x) {
	return null;
    }

    public boolean delete(BTBasicNode x, BTBasicNode p, BTBasicNode nullNode) {
	return false;
    }

    public boolean isNull() {
	return true;
    }

    /** Tausche den Nachfolger eines Elternknoten aus, der
     *  beim Aufruf auf this zeigt.
     *
     *  @param p Elternknoten von this
     *  @param newNode Austauschknoten fuer this: Wenn this
     *         der linke Nachfolger von p ist, soll newNode
     *         der neue linke Nachfolger von p werden; wenn this
     *         der rechte Nachfolger von p ist, soll newNode
     *         der neue rechte Nachfolger werden.
     *
     *  @pre   p.getLeft() == this || p.getRight() == this
     */
    public void replaceSuccessor(BTBasicNode p, BTBasicNode newNode) {
        assert p.getLeft() == this || p.getRight() == this
	    : "Illegaler Parameter p bei Aufruf von replaceSuccessor()";
	if ( this == p.getLeft() ) {
	    p.setLeft(newNode);
	}
	else {
	    p.setRight(newNode);
	}
    }



}



class BTNode extends BTBasicNode {

    TreeData data;

    public BTNode() {
	super();
	data = null;
    }

    public BTNode(TreeData x) {
	super();
	data = x;
    }

    public TreeData getData() {
	return data;
    }

    public void setData(TreeData x) {
	data = x;
    }

    public boolean isLeaf() {
	return (!(left.isNull() && right.isNull()));
    }

    @SuppressWarnings("unchecked")
    public int compareTo(BTBasicNode theNode) {
	return data.compareTo(((BTNode)theNode).getData());
    }



    public boolean delete(BTBasicNode x, BTBasicNode p, BTBasicNode nullNode) {

        // Wenn x groesser ist als this, dann im rechten
        // Nachfolger von this weitersuchen, falls x kleiner
        // ist als this, dann im linken. this wird der neue
        // Elternknoten.
	if ( compareTo(x) < 0 ) return right.delete(x,this,nullNode);
	if ( compareTo(x) > 0 ) return left.delete(x,this,nullNode);

        // Wir muessen this aus dem Baum loeschen
        if ( left.isNull() && right.isNull() ) {
            // Fall 1: this hat keine Nachfolger. Wenn
            // this der linke Nachfolger des Elternknotens p ist,
            // muss p.left auf den Null-Knoten gesetzt werden,
            // wenn this der rechte Nachfolger ist, p.right
            replaceSuccessor(p,nullNode);
	}
	else if ( left.isNull() ) {
            // Fall 2: this hat keinen linken Nachfolger.
            // Wenn p.left == this, sind aber alle rechten Nachfolger
            // von this <= p; also koennen wir p.left = this.right 
            // setzen. Wenn p.right == this, sind alle (rechten) Nachfolger
            // von this groesser als p, wir koennen also p.right = this.right
            // setzen.
            replaceSuccessor(p,right);
	}
	else if ( right.isNull() ) {
            // Fall 3: this hat keinen rechten Nachfolger.
            // Wenn p.left == this, sind aber alle linken Nachfolger
            // von this <= p; also koennen wir p.left = this.left 
            // setzen. Wenn p.right == this, sind alle (linken) Nachfolger
            // von this groesser als p, wir koennen also p.right = this.left
            // setzen.
            replaceSuccessor(p,left);
	}
	else {
            // Fall 4: this hat linke UND rechte Nachfolger.
            // Der Knoten, der gegen this ausgetauscht werden soll,
            // ist der KLEINSTE Knoten, der GROESSER als this ist.
	    BTBasicNode replacementNode = getRight();
	    BTBasicNode parent = this;
	    while ( ! replacementNode.getLeft().isNull() ) {
		parent = replacementNode;
		replacementNode = replacementNode.getLeft();
	    }
            // Der Austauschknoten hat nach Konstruktion
            // KEINEN linken Nachfolger.
            // Er bekommt die linken Nachfolger
            // von this als linke Nachfolger. Da this <= replacementNode,
            // sind auch alle linken Nachfolger von this <= replacementNode.
	    replacementNode.setLeft(left);

            // Wenn der Elternknoten von replacementNode NICHT gleich this
            // ist, bekommt er als linke Nachfolger die rechten Nachfolger
            // von replacementNode. Der neue rechte Nachfolger von
            // replacementNode wird der Elternknoten.
	    if ( parent != this ) {
		parent.setLeft(replacementNode.getRight());
		replacementNode.setRight(right);
	    }

            // Der Elternknoten von this bekommt replacementNode
            // als linken oder rechten nachfolger, jenachdem
            // ob this == p.left oder this == p.right gilt.
            replaceSuccessor(p,replacementNode);
	}

	return true;


    }

    public String toString() {

	String s;
	s = " (";
	s += left.toString();
	s += "\"" + data.toString() + "\"";
	s += right.toString();
	s += ") ";

	return s;

    }

    public BTBasicNode find(BTBasicNode x) {
	if ( compareTo(x) < 0 ) return right.find(x);
	if ( compareTo(x) > 0 ) return left.find(x);
	return this;
    }

    public boolean isNull() {
	return false;
    }

}


public class Bintree {

    BTBasicNode anchor;
    BTBasicNode nullNode;

    public Bintree() {
	anchor = new BTBasicNode();
	nullNode = new BTBasicNode();
	anchor.setLeft(nullNode);
	anchor.setRight(nullNode);
    }


    public boolean isEmpty() {
	return (anchor.getLeft() == nullNode); 
    }


    public void insert(TreeData x) {

	BTNode node = new BTNode(x);
	node.setLeft(nullNode);
	node.setRight(nullNode);
	BTBasicNode currentNode = anchor;
	BTBasicNode nextNode;

	do {

	    if ( currentNode.compareTo(node) > 0 ) {
		nextNode = currentNode.getLeft();
		if ( nextNode == nullNode ) {
		    currentNode.setLeft(node);
		}
	    }
	    else {
		nextNode = currentNode.getRight();
		if ( nextNode == nullNode ) {
		    currentNode.setRight(node);
		}
	    }

	    currentNode = nextNode;

	} while ( nextNode != nullNode );


    }

    public boolean find(TreeData x) {

	BTBasicNode n = anchor.getLeft().find(new BTNode(x));

	return (n != null);
    }

    public String toString() {
	return anchor.getLeft().toString();
    }

    public boolean delete(TreeData x) {
	return anchor.getLeft().delete(new BTNode(x),anchor,nullNode);
    }

}
